#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#define DB_DBM_HSEARCH 1

#ifdef HAVE_GDBM_H
#  include <gdbm.h>
#elif HAVE_DB_H
#  include <db.h>
#elif defined HAVE_DB4_DB_H
#  include <db4/db.h>
#elif defined HAVE_DB3_DB_H
#  include <db3/db.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include "qDecoder.h"
#include "cwb.h"
#include "cwb_db.h"

#define DB_TRY		(10)
#define BLOCK_SIZE	(1024 * 1)

static GDBM_FILE _dbf = NULL;

void DBError(void){
 const char *errstr;

 errstr = gdbm_strerror(gdbm_errno);
 qError("DataBase Access Error : %s", errstr);
}

int DBCreate(char *filename, int mode){
  GDBM_FILE newdb;

  /* ȭ ϴ Ȯ */
  if(qCheckFile(filename) == TRUE) return FALSE;

  newdb = gdbm_open(filename, BLOCK_SIZE, GDBM_WRCREAT, mode, NULL);

  if(newdb == NULL) return FALSE;

  gdbm_close(newdb);

  return TRUE;
}


/* DB READER WRITER   */
void DBOpen(char *filename, int flag){
  int i;
  
  if(_dbf != NULL)DBClose();

  for(i=0; i < DB_TRY; i++){
    _dbf = gdbm_open(filename, 0, flag, 0, NULL);
    if(_dbf != NULL)break;
    if(gdbm_errno != GDBM_CANT_BE_READER && gdbm_errno != GDBM_CANT_BE_WRITER) DBError();
    sleep(1);
  }
  if(_dbf == NULL) qError("Time Out!!! Please, try Again.");
}

void DBClose(void){
  if(_dbf == NULL) qError("There is no database to close.");
  gdbm_close(_dbf);
  _dbf = NULL;
}

char *DBFetch(char *format, ...){
  datum buf, ret;

  char key[1000 + 1];
  int status;
  va_list arglist;

  va_start(arglist, format);
  status = vsprintf(key, format, arglist);
  if(strlen(key) > 1000 || status == EOF){
    printf("DBFetch() : Key is too long or not valid");
    exit(1);
  }

  if(_dbf == NULL) qError("There is no opened database.");
  buf.dptr = key;
  buf.dsize = strlen(key) + 1;
  ret = gdbm_fetch(_dbf, buf);
  return ret.dptr;
}

int DBiFetch(char *format, ...){
  char *str;
  int  i;

  char key[1000 + 1];
  int status;
  va_list arglist;

  va_start(arglist, format);
  status = vsprintf(key, format, arglist);
  if(strlen(key) > 1000 || status == EOF){
    printf("DBiFetch() : Key is too long or not valid");
    exit(1);
  }

  str = DBFetch(key);

  if(str == NULL) return 0;
  else {
    i = atoi(str);
    free(str);
  }
  return i;
}

void DBStore(char *key, char *str, int flag){
  datum key_data, data_data;

  if(key == NULL) qError("Key string is null. Null pointer can not be stored.");
// minsu
  if(str == NULL) qError("Value string is null on key %s. Null pointer can not be stored.", key);
  //if(str == NULL) strncpy(str, "NULL", strlen("NULL"));
  if(_dbf == NULL) qError("There is no opened database.");

  key_data.dptr   = key;
  key_data.dsize  = strlen(key) + 1;
  data_data.dptr  = str;
  data_data.dsize = strlen(str) + 1;

  if(gdbm_store(_dbf, key_data, data_data, flag) != 0)DBError();  
}

void DBiStore(char *key, int value, int flag){
  char buf[0xFF];

  sprintf(buf, "%d", value);
  DBStore(key, buf, flag);
}

int DBDelete(char *format, ...){
  datum buf;
  int ret;

  char key[1000 + 1];
  int status;
  va_list arglist;

  va_start(arglist, format);
  status = vsprintf(key, format, arglist);
  if(strlen(key) > 1000 || status == EOF){
    printf("DBDelete() : Key is too long or not valid");
    exit(1);
  }

  if(_dbf == NULL) qError("There is no database to fetch.");
  buf.dptr  = key;
  buf.dsize = strlen(key) + 1;
  ret = gdbm_delete(_dbf, buf);
  return ret;
}
