#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#define DB_DBM_HSEARCH 1

#ifdef HAVE_GDBM_H
#  include <gdbm.h>
#elif HAVE_DB_H
#  include <db.h>
#elif defined HAVE_DB4_DB_H
#  include <db4/db.h>
#elif defined HAVE_DB3_DB_H
#  include <db3/db.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "qDecoder.h"
#include "cwb.h"
#include "cwb_db.h"
#include "cwb_list.h"

/********* Index Sturct *************/
typedef struct D_INDEX {
  int num;
  int thread;
}D_INDEX;

typedef struct T_INDEX {
  int num;
  int depth;
}T_INDEX;

static D_INDEX *dIndex = NULL;
static T_INDEX *tIndex = NULL;

static int totalArticle = 0;

/********* Index Stack *************/
#define STACK_SIZE	(100)  /*  Խù ķ Ҽ ִ  ټ */
typedef struct INDEX_STACK {
  int num1[STACK_SIZE+1];
  int num2[STACK_SIZE+1];
}INDEX_STACK;

static INDEX_STACK stack;
static int sp = 0;

/********** Common Functions **************/
int  main_index(cDB *db, DBConfig *conf);
int  get_index(int num, int ftype, char *fval);
void adjust_index(void);
int  findfval(int num, int ftype, char *fval);

void date_index(void);

void thread_index(void);
int  findThread(int num, int upper);
int  insertIndex(int num, int depth);
int  pushIndex(int num, int i);
int  popIndex(int *num, int *i);

void print_index(Q_Entry *sysp, SysConfig *sys, cDB *db, DBConfig *conf, int tArticle);
void shotSearch(Q_Entry *sysp, SysConfig *sys, char *dbname, char *language, int ftype, char *fval);
int  getArticleInfo(int num, DBList *head);
void freeArticleInfo(DBList *head);

void strbig(char *str);
int  strfind(char *orgstr, char *tokstr);

/*********** Main ***************************/
void cwb_list(Q_Entry *sysp, SysConfig *sys, cDB *db, DBConfig *conf){
  char buf[0xFF];
  int  tArticle;

  /* DB Open */
  sprintf(buf, "%s/%s.%s", sys->DataDirectory, db->db, sys->DataExtension);
  DBOpen(buf, CWDB_READ);

  tArticle = main_index(db, conf);

  print_index(sysp, sys, db, conf, tArticle);

  /* DB Close */
  DBClose();
}


/* Խù Ʒ ȣ ȹѴ, Լ     
   cwb_read()  ȣȴ. */
void findUPDNnum(char *dbpath, cDB *db, DBConfig *conf, int *upNum, char *upSubject, int *dnNum, char *dnSubject) {
  int  i;

  /* DB Open */
  DBOpen(dbpath, CWDB_READ);

  main_index(db, conf);

  for(i = 0; tIndex[i].num != 0; i++) {
    if(tIndex[i].num == db->num) {
      if(i - 1 < 0) *upNum = 0;
      else *upNum = tIndex[i-1].num;
      *dnNum = tIndex[i+1].num;
    }
  }

  if(*upNum == 0) strcpy(upSubject, "");
  else strcpy(upSubject, qRemoveSpace(DBFetch("%d.Subject", *upNum)));
  if(*dnNum == 0) strcpy(dnSubject, "");
  else strcpy(dnSubject, qRemoveSpace(DBFetch("%d.Subject", *dnNum)));

  /* DB Close */
  DBClose();
}


/*************************************
** DB  Լ ȣϸ,  Ȳ Ͽ
**  Ʈ ,  ټ Ѵ.
** ϰ ϰ˻  Խù ̸
** װ͸ Ʈ ȴ.
**  ε Լ ȣϴ ̴.
**************************************/
int main_index(cDB *db, DBConfig *conf){
  int  tArticle;

  tArticle = get_index(conf->LastNum, db->ftype, db->fval);
  if(tArticle == -1) qError("There is a fatal error in database(%s)<br>You should better call the administrator.", db->db);

  adjust_index();

  if(!strcmp(conf->ListOrder, "date")) date_index();
  else if(!strcmp(conf->ListOrder, "thread")) thread_index();
  else qError("There is a fatal error in database(%s)<br>You should better call the administrator.", db->db);

  return tArticle;
}

/*************************************
** DB  , ϰ˻Ͽ ɷ Խù ε̵ȴ.
** [0] -> 10, 0 (num, thread)
** [1] ->  9, 1
** [2] ->  8, 0
** [n] ->  0, 0
** Խù  Ѵ.  -1
**************************************/
int get_index(int num, int ftype, char *fval) {
  int i, j;
  char *str;

  if(num < 0) return -1;

  dIndex = (D_INDEX *)malloc(sizeof(D_INDEX) * (num + 1));
  if(dIndex == NULL) qError("Memory allocation fail.");
  tIndex = (T_INDEX *)malloc(sizeof(T_INDEX) * (num + 1));
  if(tIndex == NULL) qError("Memory allocation fail.");

  /* DB Fetch */
  for(i = num, j = 0; i > 0; i--){
    str = DBFetch("%d.Thread", i);

    /* Thread DB  , Խù Ǿٰ Ѵ */
    if(str != NULL) {  /* Խù  */
      totalArticle++;  /* ü ù  */
      /*  ˻ */
      if(findfval(i, ftype, fval) == TRUE) {
        dIndex[j].num    = i;
        dIndex[j].thread = atoi(str);
        j++;
      }
      free(str);
    }
  }
  dIndex[j].num = dIndex[j].thread = 0;

  return j;
}

/*************************************
** index Ѵ. ϰ˻縦 ģ ε 𾲷尡  ʴ
** Խù  ִ.
**  尡 ε  ȣ 0  Ͽ Thread list order
** ϶ Խù Ȯεǵ Ѵ.
** ǻ ʿ,  reference thread  Ѵٸ
**  Լ ̸ Ѵ.
**************************************/
void adjust_index(void) {
  int i, j;

  /* DB Fetch */
  for(i = 0; dIndex[i].num != 0; i++){
    if(dIndex[i].thread == 0) continue;

    /*  尡 ϴ ˻ */
    for(j = i + 1; dIndex[j].num != 0; j++) {
      if(dIndex[j].num == dIndex[i].thread) break;
    }
    /*  Խù  */
    if(dIndex[j].num == 0) dIndex[i].thread = 0;
  }
}

/* ϰ˻Ͽ Խù մ ǺѴ.*/
int findfval(int num, int ftype, char *fval) {
  char *name, *subject, *text;
  int  cf_name, cf_subject, cf_text;
  int  findflag;

  /* ˻ڿ  ãɷ  */
  if(!strcmp(fval, "") || ftype == 0) return TRUE;
  
  /* ʱȭ */
  cf_name = cf_subject = cf_text = FALSE;
  name    = subject    = text    = NULL;

  /* ˻׸ Ǻ */
  if((ftype & FNAME)    != 0) cf_name    = TRUE;
  if((ftype & FSUBJECT) != 0) cf_subject = TRUE;
  if((ftype & FTEXT)    != 0) cf_text    = TRUE;

  /* Fetch */
  if(cf_name    == TRUE) name    = DBFetch("%d.Name", num);
  if(cf_subject == TRUE) subject = DBFetch("%d.Subject", num);
  if(cf_text    == TRUE) text    = DBFetch("%d.Text", num);

  /* NULL ƴѰ ã´*/
  if(strfind(name, fval)    == TRUE) findflag = TRUE;
  else if(strfind(subject, fval) == TRUE) findflag = TRUE;
  else if(strfind(text, fval)    == TRUE) findflag = TRUE;
  else findflag = FALSE;

  /* Free */
  if(name    != NULL) free(name);
  if(subject != NULL) free(subject);
  if(text    != NULL) free(text);

  return findflag;
}

/* Date List */
void date_index(void) {
  int i;
  for(i = 0; dIndex[i].num != 0; i++){
    insertIndex(dIndex[i].num, 0);
  }
  insertIndex(0, 0);
}

/* Thread List */
void thread_index(void) {
  int di;  /* dIndex index  */

  int threadNum;  /*  Խù  Խù ȣ */
  int threadI;    /*  Խù dIndex  */
  int depth;
  int exit;       /*  Ż  */

  for (di = 0; dIndex[di].num != 0; di++){
    /* thread 0 Խù ã´. */
    if(dIndex[di].thread != 0) continue;

    /* ó ⿡ , ֻ Խù ã ǹѴ */

    /* tIndex depth 0 Խù ߰Ѵ. */
    insertIndex(dIndex[di].num, 0);

    /* ã   num , upper  Ѵ */
    threadNum = dIndex[di].num, threadI = di;

    exit = FALSE;
    do {
      /* 带 ã´ */
      threadI = findThread(threadNum, threadI);

      if(threadI >= 0) { /* 带 ã */
        /* Stack   num ã  index ǪѴ */
        depth = pushIndex(threadNum, threadI);

        /* Խù ߰ */
        insertIndex(dIndex[threadI].num, depth);

        /*  ã 尪 Ѵ */
        threadNum = dIndex[threadI].num;
      }
      else { /*   ã */
        if(popIndex(&threadNum, &threadI) == FALSE) exit = TRUE;
      }
    }while(exit == FALSE);
  }
  insertIndex(dIndex[di].num, 0);
}

/* num  ϴ  Խù dIndex[upper-1]  ã´ */
int findThread(int num, int upper){
  int i;

  for(i = upper - 1; i >= 0; i--){
    if(dIndex[i].thread == num) return i;
  }

  return -1;
}

/* tIndex Խù ߰Ѵ. Ͱ Խù   */
int insertIndex(int num, int depth) {
  static int Ip = 0;

  tIndex[Ip].num   = num;
  tIndex[Ip].depth = depth;
  Ip++;

  return Ip;
}

/*   num,  ڽ Index ǪѴ */
/*      */
int pushIndex(int num, int i) {
  if(sp > STACK_SIZE) return FALSE;
  stack.num1[sp] = num;
  stack.num2[sp] = i;
  sp++;
  return sp;
}

int popIndex(int *num, int *i) {
  if(sp == 0) return FALSE;
  sp--;
  *num = stack.num1[sp];
  *i   = stack.num2[sp];
  return TRUE;
}

void print_index(Q_Entry *sysp, SysConfig *sys, cDB *db, DBConfig *conf, int tArticle){
  DBList head;
  char *Number, *Type, *Subject, *Name, *Date, *Access, *Lines, *Checkbox;

  char fileicon[0xff];

  int startNum,  endNum, pages;
  int startPage, endPage;

  int  i, status, virtualnum, nbspnum;


  char status_msg[0xff]; /* ũ ϴ ¹ٿ Ÿ ޽ */
  char *msg, *msg1; /* ޽ */

  if(!strcmp(conf->Language, "korean")) {
    Number  = "ȣ";
    Type    = "&copy;";
    Subject = " ";
    Name    = " ";
    Date    = "  ";
    Access  = " ȸ"; 
    Lines   = " ";
    Checkbox= "";
  }
  else {
    Number  = "No";
    Type    = "&copy;";
    Subject = "S u b j e c t";
    Name    = "Name";
    Date    = "Date";
    Access  = "Access";
    Lines   = "Lines";
    Checkbox= "";
  }

  if(qfValue(sysp, "TextNumber") != NULL) Number = qfValue(sysp, "TextNumber");
  if(qfValue(sysp, "TextType") != NULL) Type = qfValue(sysp, "TextType");
  if(qfValue(sysp, "TextSubject") != NULL) Subject = qfValue(sysp, "TextSubject");
  if(qfValue(sysp, "TextName") != NULL) Name = qfValue(sysp, "TextName");
  if(qfValue(sysp, "TextDate") != NULL) Date = qfValue(sysp, "TextDate");
  if(qfValue(sysp, "TextAccess") != NULL) Access = qfValue(sysp, "TextAccess");
  if(qfValue(sysp, "TextLines") != NULL) Lines = qfValue(sysp, "TextLines");
  if(qfValue(sysp, "TextCheckbox") != NULL) Lines = qfValue(sysp, "TextCheckbox");

  /* Page   */
  if(tArticle == 0) pages = 0;
  else {
    int pof;

    if((tArticle % conf->ArticlesPerPage) == 0) pof = 0;
    else pof = 1;

    pages = (tArticle / conf->ArticlesPerPage) + pof;
  }

  /* ũ Page   */
  if(pages == 0) startPage = endPage = 0;
  else {
    int pof;

    if((db->page % atoi(sys->PageLinkNum)) == 0) pof = 0;
    else pof = 1;

    endPage   = ((db->page / atoi(sys->PageLinkNum)) + pof) * atoi(sys->PageLinkNum);
    startPage = endPage - (atoi(sys->PageLinkNum) - 1);
  }

  /*    */
  if(db->page > pages) db->page = pages;

  /*  Index ȣ  */
  if(db->page == 0) startNum = endNum = 0;
  else {
    startNum = (db->page - 1) * conf->ArticlesPerPage;
    endNum   = startNum + (conf->ArticlesPerPage - 1 );
  }

  /* Header  */
  headHtml(sys, conf);

  printf("\n<!-- HR Line -->\n");
  printf("<table border=0 width=80%%>\n");
  printf("  <tr><td><hr size=1 noshade></td></tr>\n");
  printf("</table>\n");

  /* DB Information */
  printf("\n<!-- DB Information -->\n");
  printf("<table border=0 width=80%%><tr>\n");
  if(strcmp(sys->HiddenAdminLink, "true") != 0) {
    strcpy(status_msg, "Change database configurations, only administrator can access");
    printf("  <td align=left><a href='%s?db=%s&mode=admin&backdepth=1' onMouseOver=\"window.status=('%s'); return true;\"><img src='%s' border=0 alt='Administration'></a></td>\n", sys->ProgramName, db->db, status_msg, sys->IconAdmin);
  }
  if(qfValue(sysp, "TextTotalArticles") != NULL) msg = qfValue(sysp, "TextTotalArticles");
  else msg="Total Articles";
  if(qfValue(sysp, "TextTotalPage") != NULL) msg1 = qfValue(sysp, "TextTotalPage");
  else msg1="Page";

  printf("  <td align=right><font size='%d' face='Sans-serif'>%s : <font color=darkgreen>%d</font> / <font color=darkblue>%d</font> , %s : <font color=red>%d</font> / <font color=darkblue>%d</font>\n", atoi(conf->ListFontSize) - 1, msg, tArticle, totalArticle, msg1, db->page, pages);
  if(qfValue(sysp, "UseFeed") != NULL) {
    char *feedicon;
    feedicon = qfValue(sysp, "IconFeed");
    if (feedicon == NULL) feedicon = "XML";
    printf("<a href=\"%s?db=%s&amp;mode=feed\">%s</a>\n",sys->ProgramName,db->db,feedicon);
  }
  printf("</td></tr></table>\n");

  /* Body */
  printf("\n<!-- List Body -->\n");

  if(!strcmp(conf->AdminCheckbox, "true")) {
    printf("<form id='listForm' method='get' action='%s'>\n", sys->ProgramName);
    printf("  <input type='hidden' name='db' value='%s'>\n", db->db);
    printf("  <input type='hidden' name='page' value='%d'>\n", db->page);
    printf("  <input type='hidden' name='ftype' value='%d'>\n", db->ftype);
    printf("  <input type='hidden' name='fval' value='%s'>\n", db->fencval);
    printf("  <input type='hidden' name='backdepth' value='%d'>\n", db->backdepth + 1);
    printf("  <input type='hidden' name='mode' value='delete'>\n");
  }

  printf("<table border='%s' width=80%%>\n", conf->ListBorderSize);
  printf("  <tr bgcolor='%s'>\n", conf->ListTitleBgColor);
  if(!strcmp(conf->ArticleNumber, "true")) printf("    <th width='5%%'><font size='%s' face='Sans-serif'>%s</font></th>\n", conf->ListFontSize, Number);
  if(!strcmp(conf->NameFirst, "true"))     printf("    <th width='12%%'><font size='%s' face='Sans-serif' color='%s'><nobr>%s</nobr></font></th>\n", conf->ListFontSize, conf->ListTitleFontColor, Name);
  if(!strcmp(conf->AdminCheckbox, "true")) printf("    <th width='2%%'><font size='%s' face='Sans-serif' color='%s'>%s</font></th>\n", conf->ListFontSize, conf->ListTitleFontColor, Checkbox);
  if(!strcmp(conf->TypeIcon, "true"))      printf("    <th width='2%%'><font size='%s' face='Sans-serif' color='%s'><nobr>%s</nobr></font></th>\n", conf->ListFontSize, conf->ListTitleFontColor, Type);
  printf("    <th><font size='%s' face='Sans-serif' color='%s'><nobr>%s</nobr></font></th>\n", conf->ListFontSize, conf->ListTitleFontColor, Subject);
  if(!strcmp(conf->NameFirst, "false"))    printf("    <th width='12%%'><font size='%s' face='Sans-serif' color='%s'><nobr>%s</nobr></font></th>\n", conf->ListFontSize, conf->ListTitleFontColor, Name);
  printf("    <th width='10%%'><font size='%s' face='Sans-serif' color='%s'><nobr>%s</nobr></font></th>\n", conf->ListFontSize, conf->ListTitleFontColor, Date);
  printf("    <th width='5%%'><font size='%s' face='Sans-serif' color='%s'><nobr>%s</nobr></font></th>\n", conf->ListFontSize, conf->ListTitleFontColor, Access);
  printf("  </tr>\n\n");


  /* Ϻ  thread϶ ñ۵鶧  Խù ȣ
     ִ.  ̶ üù  ȣ Ͽ
     Ϸ ټֵ Virtual Num ɼ ־.
         Խù ǰų,  ñ ۼǸ
     Խù ȣ  Ѵٴ ̴.
     ׷, ñ϶ Խùȣ ϴ Ķ ̻ Ͽ
     ܰ  .
     忡 ڰ ̸ Ͽ,  ȣ ǥ ȣ
     ǥ Ѵ. */

  virtualnum = tArticle - ((db->page - 1) * conf->ArticlesPerPage);

  /* tIndex   0 ʵ尡 ԵǾ ִ */
  for(i = startNum; (i <= endNum) && (tIndex[i].num != 0); i++, virtualnum--){
    status = getArticleInfo(tIndex[i].num, &head);

    printf("  <tr bgcolor='%s'><!-- Article %d -->\n", conf->ListArticleBgColor, tIndex[i].num);

    /* Article Number */
    if(!strcmp(conf->ArticleNumber, "true")) {
      printf("    <td align=center><font size='%s' face='Sans-serif' color='%s'>", conf->ListFontSize, conf->ListArticleFontColor);
      if(!strcmp(conf->VirtualNumber, "true")) printf("%d", virtualnum);
      else printf("%d", tIndex[i].num);
      printf("</font></td>\n");
    }

    /* Is it brocken? */
    if(status == FALSE) { /* Brocken Article*/
      if(!strcmp(conf->NameFirst, "true")) printf("    <td>&nbsp;</td>\n");
      if(!strcmp(conf->TypeIcon, "true")) printf("    <td>&nbsp;</td>\n");
      printf("    <th><font size='%s' face='Sans-serif' color='red'>Brocken Article : %d</font></th>\n", conf->ListFontSize, tIndex[i].num);
      if(!strcmp(conf->NameFirst, "false")) printf("    <td>&nbsp;</td>\n");
      printf("    <td>&nbsp;</td>\n");
      printf("    <td>&nbsp;</td>\n");
    }
    else {/* Fine Article */

      /* Name First*/
      if(!strcmp(conf->NameFirst, "true")) {
        printf("    <td><font size='%s' face='Sans-serif' color='%s'><center><nobr>", conf->ListFontSize, conf->ListArticleFontColor);
        if(!strcmp(head.Email, "")) highlightStr(1, head.Name, db->fval);
        else {
          sprintf(status_msg, "Send mail to %s (%s)", head.Name, head.Email);
          printf("<a href='mailto:%s <%s>' onMouseOver=\"window.status=('%s'); return true;\">", head.Name, head.Email, status_msg);
          highlightStr(1, head.Name, db->fval);
          printf("</a>");
        }
        printf("</nobr></center></font></td>\n");
      }

      /* Admin Checkbox */
      if(!strcmp(conf->AdminCheckbox, "true")) {
        printf("    <td align=center>");
        printf("<input type='checkbox' name='num' value='%d'>", tIndex[i].num );
        printf("</td>\n");
      }

      /* File Type Icon*/
      if(!strcmp(conf->TypeIcon, "true")) {
        printf("    <td align=center>");
        switch(findExtIcon(sysp, sys, head.Filename, fileicon)) {
          case TYPE_UNLINK : {
            printf("<img src='%s' width=%s height=%s border=0>", fileicon, sys->IconWidth, sys->IconHeight);
            break;
          }
          default : {
            sprintf(status_msg, "Download : %s (%d Kbytes)", head.Filename, atoi(head.Filesize) / 1024);
            printf("<a href='%s/%s/%s' onMouseOver=\"window.status=('%s'); return true;\"><img src='%s' width=%s height=%s border=0 alt='%s'></a>", sys->DataDirectory, db->db, head.Filename, status_msg, fileicon, sys->IconWidth, sys->IconHeight, status_msg);
            break;
          }
        }
        printf("</td>\n");
      }

      /* Subject */
      printf("    <td align=left><font size='%s' face='Sans-serif' color='%s'><nobr>&nbsp;&nbsp;", conf->ListFontSize, conf->ListArticleFontColor);


      /*   ִ´. */
      for(nbspnum = 0; nbspnum < tIndex[i].depth; nbspnum++) {
        printf("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
      }

      /* ICON OPEN, CLOSE */
      if(!strcmp(conf->FolderIcon, "true")) {
        if(tIndex[i+1].depth > tIndex[i].depth) printf("<img src='%s' width=%s height=%s border=0>", sys->IconOpen, sys->IconWidth, sys->IconHeight);
        else printf("<img src='%s' width=%s height=%s border=0>", sys->IconClose, sys->IconWidth, sys->IconHeight);
        printf("&nbsp;&nbsp;");
      }
      sprintf(status_msg, "Physical number %d , Virtual number %d, Included %d lines", head.Num, virtualnum, head.Lines);
      printf("<a href='%s?db=%s&mode=read&num=%d&page=%d&ftype=%d&fval=%s&backdepth=1' onMouseOver=\"window.status=('%s'); return true;\">", sys->ProgramName, db->db, tIndex[i].num, db->page, db->ftype, db->fencval, status_msg);
      highlightStr(11, head.Subject, db->fval);
      printf("</a>");

      printf("</nobr></font></td>\n");

      /* Name Second */
      if(!strcmp(conf->NameFirst, "false")) {
        printf("    <td><font size='%s' face='Sans-serif' color='%s'><center><nobr>", conf->ListFontSize, conf->ListArticleFontColor);
        if(!strcmp(head.Email, "")) highlightStr(1, head.Name, db->fval);
        else {
          sprintf(status_msg, "Send mail to %s (%s)", head.Name, head.Email);
          printf("<a href='mailto:%s <%s>' onMouseOver=\"window.status=('%s'); return true;\">", head.Name, head.Email, status_msg);
          highlightStr(1, head.Name, db->fval);
          printf("</a>");
        }
        printf("</nobr></center></font></td>\n");
      }

      /* Date, Access */
      printf("    <td align=center><font size='%s' face='Sans-serif' color='%s'>%02d/%02d/%02d</font></td>\n", conf->ListFontSize, conf->ListArticleFontColor, head.Year, head.Mon, head.Day);
      printf("    <td align=center><font size='%s' face='Sans-serif' color='%s'>%d</font></td>\n", conf->ListFontSize, conf->ListArticleFontColor, head.Access);
    }
    printf("  </tr>\n");
    freeArticleInfo(&head);
  }

  /* Delete Articles Link */
  if(!strcmp(conf->AdminCheckbox, "true")) {
    printf("<tr>");
    if(!strcmp(conf->ArticleNumber, "true")) printf("<td></td>");
    printf("<td align=center>");
    printf("<script type='text/javascript'>\n"
"/*<![CDATA[*/\n"
"  function send_list(formname,mode) {\n"
"    var tmp=\"\";var i, chk=false;\n"
"\n"
"    form=document.getElementById(formname);\n"
"\n"
"    for(i=0;i< form.length;i++) {\n"
"       if(form[i].type!=\"checkbox\") continue;\n"
"       if(form[i].checked) {\n"
"          chk=true;break;\n"
"       }\n"
"    }\n"
"    if(chk==true) {\n"
"       form.submit();\n"
"       return false;\n"
"    }\n"
"    alert (\"Please check article numbers.\");\n"
"    return false;\n"
"  }\n"
"/*]]>*/\n"
"</script>\n");
    printf("      <a href='%s?db=%s' onclick=\"return send_list('listForm','delete');\" onfocus=\"blur()\" ><img src='%s' border=0 alt='Delete'></a>\n", sys->ProgramName, db->db, sys->IconDelete);
    printf("</td></tr>\n");
  }
  printf("</table>\n");
  if(!strcmp(conf->AdminCheckbox, "true")) printf("</form>\n");

  /* Page Link */
  printf("\n<!-- Page Link -->\n");
  printf("<table border=0 width=80%%>\n");
  printf("  <tr><td align=center><font size='%d' face='arial'>\n", atoi(conf->ListFontSize) - 1 );

  if(startPage > 1) {
    strcpy(status_msg, "Jump to first page");
    printf("    <a href='%s?db=%s&page=%d&ftype=%d&fval=%s' onMouseOver=\"window.status=('%s'); return true;\">[First]</a>\n", sys->ProgramName, db->db, 1, db->ftype, db->fencval, status_msg);
    strcpy(status_msg, "Jump to previous page");
    printf("    <a href='%s?db=%s&page=%d&ftype=%d&fval=%s' onMouseOver=\"window.status=('%s'); return true;\">[Prev]</a> -\n", sys->ProgramName, db->db, startPage-1, db->ftype, db->fencval, status_msg);
  }
  for(i = startPage; (i <= endPage) && (i <= pages) && (i > 0); i++) {
    if(i == db->page) {
      printf("    [<font color=red>%d</font>]\n", i);
    }
    else {
      sprintf(status_msg, "Jump to page %d", i);
      printf("    <a href='%s?db=%s&page=%d&ftype=%d&fval=%s' onMouseOver=\"window.status=('%s'); return true;\">[%d]</a>\n", sys->ProgramName, db->db, i, db->ftype, db->fencval, status_msg, i);
    }
  }
  if(endPage < pages) {
    strcpy(status_msg, "Jump to next page");
    printf("  - <a href='%s?db=%s&page=%d&ftype=%d&fval=%s' onMouseOver=\"window.status=('%s'); return true;\">[Next]</a>\n", sys->ProgramName, db->db, endPage+1, db->ftype, db->fencval, status_msg);
    if(endPage + 1 < pages) {
      strcpy(status_msg, "Jump to last page");
      printf("    <a href='%s?db=%s&page=%d&ftype=%d&fval=%s' onMouseOver=\"window.status=('%s'); return true;\">[Last]</a>\n", sys->ProgramName, db->db, pages, db->ftype, db->fencval, status_msg);
    }
  }
  printf("  </font></td></tr>\n");
  printf("</table>\n");

  printf("\n<!-- HR Line -->\n");
  printf("<table border=0 width=80%%>\n");
  printf("  <tr><td><hr size=1 noshade></td></tr>\n");
  printf("</table>\n");

  /* ICON Link */
  printf("\n<!-- ICON Link -->\n");
  printf("<table border=0 width=80%%><tr>\n");
  printf("  <td align=left>\n");
  /* HomePage */
  if(strcmp(conf->HomeUrl, "")) {
    printf("    <a href='%s' target='%s'><img src='%s' border=0 alt='HomePage'></a>\n", conf->HomeUrl, conf->HomeTarget, sys->IconHome);
  }
  /* BackPage */
  if(strcmp(conf->BackUrl, "")) {
    printf("    <a href='%s' target='%s'><img src='%s' border=0 alt='Backpage'></a>\n", conf->BackUrl, conf->BackTarget, sys->IconBack);
  }
  /* Help */
  if(strcmp(sys->HiddenHelpLink, "true") != 0) {
    strcpy(status_msg, "Help message about this billboard");
    printf("    <a href='%s' onMouseOver=\"window.status=('%s'); return true;\"><img src='%s' border=0 alt='Help'></a>\n", sys->Help, status_msg, sys->IconHelp);
  }
  printf("  </td>\n");

  /* Search */
  printf("\n<!-- Search -->\n");
  printf("  <td align=center>\n");
  if(!strcmp(conf->HideSearch, "true")) printf("&nbsp;");
  else shotSearch(sysp, sys, db->db, conf->Language, db->ftype, db->fval);
  printf("  </td>\n");

  printf("\n<!-- ICON Link -->\n");
  printf("  <td align=right>\n");
  /* Up */
  if(db->page > 1) {
    strcpy(status_msg, "Previous page");
    printf("    <a href='%s?db=%s&mode=list&page=%d&ftype=%d&fval=%s' onMouseOver=\"window.status=('%s'); return true;\"><img src='%s' border=0 alt='Backward'></a>\n", sys->ProgramName, db->db, db->page-1, db->ftype, db->fencval, status_msg, sys->IconPrev);
  }
  /* Down */
  if(db->page < pages) {
    strcpy(status_msg, "Next page");
    printf("    <a href='%s?db=%s&mode=list&page=%d&ftype=%d&fval=%s' onMouseOver=\"window.status=('%s'); return true;\"><img src='%s' border=0 alt='forward'></a>\n", sys->ProgramName, db->db, db->page+1, db->ftype, db->fencval, status_msg, sys->IconNext);
  }
  /* Write */
  strcpy(status_msg, "Write message");
  printf("    <a href='%s?db=%s&mode=write&backdepth=1' onMouseOver=\"window.status=('%s'); return true;\"><img src='%s' border=0 alt='Post'></a>\n", sys->ProgramName, db->db, status_msg, sys->IconWrite);
  /* Reload List */
  strcpy(status_msg, "Reload list");
  printf("    <a href='%s?db=%s' onMouseOver=\"window.status=('%s'); return true;\"><img src='%s' border=0 alt='Reload'></a>\n", sys->ProgramName, db->db, status_msg, sys->IconReload);

  printf("  </td>\n");
  printf("</tr></table>\n");

  tailHtml(sys, conf);
}

void shotSearch(Q_Entry *sysp, SysConfig *sys, char *dbname, char *language, int ftype, char *fval){
  char *Name, *Subject, *Text, *Search;
  char *checkflag;

  if(!strcmp(language, "korean")) {
    Name    = "";
    Subject = "";
    Text    = "";
    Search  = "˻";
  }
  else {
    Name    = "Name";
    Subject = "Subject";
    Text    = "Text";
    Search  = "Go";
  }

  if(qfValue(sysp, "SearchName") != NULL) Name = qfValue(sysp, "SearchName");
  if(qfValue(sysp, "SearchSubject") != NULL) Subject = qfValue(sysp, "SearchSubject");
  if(qfValue(sysp, "SearchText") != NULL) Text = qfValue(sysp, "SearchText");
  if(qfValue(sysp, "SearchGo") != NULL) Search = qfValue(sysp, "SearchGo");

  printf("<form method='post' action='%s'>\n", sys->ProgramName);
  printf("  <input type='hidden' name='db'   value='%s'>\n", dbname);
  printf("  <input type='hidden' name='mode' value='list'>\n");


  printf("<table border=0 cellpadding=0 cellspacing=0>\n");
  printf("  <tr>\n");
  printf("    <td align=right><font size=2 face='Sans-serif'>\n");

  if((ftype & (int)FNAME)    != 0) checkflag = "checked";
  else                             checkflag = "";
  printf("      %s <input type='checkbox' name='fname'    value='checked' %s><br>\n", Name, checkflag);

  if((ftype & (int)FSUBJECT) != 0) checkflag = "checked";
  else                             checkflag = "";
  printf("      %s <input type='checkbox' name='fsubject' value='checked' %s><br>\n", Subject, checkflag);

  if((ftype & (int)FTEXT)    != 0) checkflag = "checked";
  else                             checkflag = "";
  printf("      %s <input type='checkbox' name='ftext'    value='checked' %s><br>\n", Text, checkflag);

  printf("    </font></td>\n");

  printf("    <td align=left>\n");
  printf("      <font size=3><input type='text' name='fval' value='%s' size=14 maxlength=20></font>\n", fval);
  printf("      <font size=2 face='Sans-serif'><input type='submit' value='%s'></font>\n", Search);
  printf("    </td>\n");
  printf("  </tr>\n");
  printf("</table>\n");
  printf("</form>\n");
}

int getArticleInfo(int num, DBList *head){
  head->Num = num;

  /* freeҶ   ̸ NULL ʱȭ Ѵ */
  head->Name    = NULL;
  head->Email   = NULL;
  head->Subject = NULL;
  head->Date    = NULL;

  /* Fetch */
  if((head->Name     = qRemoveSpace(DBFetch("%d.Name", num))) == NULL) return FALSE;
  if((head->Email    = qRemoveSpace(DBFetch("%d.Email", num))) == NULL) return FALSE;
  if((head->Subject  = qRemoveSpace(DBFetch("%d.Subject", num))) == NULL) return FALSE;
  if((head->Date     = DBFetch("%d.Date", num)) == NULL) return FALSE;

  if((head->Filename = DBFetch("%d.Filename", num)) == NULL) head->Filename = "";
  if((head->Filesize = DBFetch("%d.Filesize", num)) == NULL) head->Filesize = "";

  /* YYYYMMDD иѴ */
  sscanf(head->Date, "%4d%2d%2d", &head->Year, &head->Mon, &head->Day);

  /* ⵵ տ 2ڸ Ѵ */
  for(; head->Year >= 100; head->Year -= 100);

  head->Lines  = DBiFetch("%d.Lines", num);
  head->Access = DBiFetch("%d.Access", num);
  return TRUE;
}

void freeArticleInfo(DBList *head){
  if(head->Name    != NULL) free(head->Name);
  if(head->Email   != NULL) free(head->Email);
  if(head->Subject != NULL) free(head->Subject);
  if(head->Date    != NULL) free(head->Date);
}

void strbig(char *str){
  int i = 0;

  while(str[i]){
    if(str[i] >= 97 && str[i] <= 122)str[i] -= 32;
    i++;
  }
}

int strfind(char *orgstr, char *tokstr){
  char *org, *tok;
  int  findflag;

  if(orgstr == NULL || tokstr == NULL) return FALSE;

  if((org = (char *)malloc(strlen(orgstr) + 1)) == NULL) return FALSE;
  if((tok = (char *)malloc(strlen(tokstr) + 1)) == NULL) {
    free(org);
    return FALSE;
  }

  strcpy(org, orgstr), strcpy(tok, tokstr);

  strbig(org), strbig(tok);

  if(strstr(org, tok) == NULL) findflag = FALSE;
  else                         findflag = TRUE;

  free(org), free(tok);

  return findflag;
}


/* ϸ ָ ׿شϴ ¸ ϸ, Ͽ شϴ ICON Ѱش */
int findExtIcon(Q_Entry *sysp, SysConfig *sys, char *filename, char *icon) {
  char ext[0xff], buf[0xff];
  int  i, finded;

  /* ϸ Ȯڸ и */
  for(strcpy(ext, filename), finded = FALSE, i = strlen(filename) - 1; i >= 0; i--) {
    if((ext[i] == '.') || (finded == TRUE)) {
      finded = TRUE;
      ext[i] = ' ';
    }
    qRemoveSpace(ext);
    strbig(ext); /* Ȯڸ 빮ڷ ȯ */
  }
  if (finded == FALSE) {
    strcpy(icon, sys->IconTypeDefault);
    return TYPE_UNLINK;
  }

  /* Check general binary */
  sprintf(buf, "IconType-%s", ext);
  if(qfValue(sysp, buf) != NULL) {
    strcpy(icon, qfValue(sysp, buf));
    return TYPE_LINK;
  }

  /* Check general text */
  sprintf(buf, "IconTypeText-%s", ext);
  if(qfValue(sysp, buf) != NULL) {
    strcpy(icon, qfValue(sysp, buf));
    return TYPE_TEXT;
  }

  /* Check general image */
  sprintf(buf, "IconTypeImage-%s", ext);
  if(qfValue(sysp, buf) != NULL) {
    strcpy(icon, qfValue(sysp, buf));
    return TYPE_IMAGE;
  }

  /* Check general sound */
  sprintf(buf, "IconTypeSound-%s", ext);
  if(qfValue(sysp, buf) != NULL) {
    strcpy(icon, qfValue(sysp, buf));
    return TYPE_SOUND;
  }

  /* Check general movie */
  sprintf(buf, "IconTypeMovie-%s", ext);
  if(qfValue(sysp, buf) != NULL) {
    strcpy(icon, qfValue(sysp, buf));
    return TYPE_MOVIE;
  }

  /* Not found */
  strcpy(icon, sys->IconTypeUnknown);
  return TYPE_LINK;
}

void highlightStr(int mode, char *orgstr, char *orgtok) {
  char *str, *tok, *buf;
  char *freestr, *freetok;
  char *findp;
  int  i;

  if(!strcmp(orgtok, "")) {
    qPuts(mode, orgstr);
    return;
  }

  /* 빮ڷ ȯ */
  if((str = freestr = (char *)malloc(strlen(orgstr) + 1)) == NULL) qError("Memory allocation fail.");
  if((tok = freetok = (char *)malloc(strlen(orgtok) + 1)) == NULL) qError("Memory allocation fail.");

  strcpy(str, orgstr), strcpy(tok, orgtok);
  strbig(str), strbig(tok);

  /*   */
  if((buf = (char *)malloc(strlen(orgstr) + 1)) == NULL) qError("Memory allocation fail.");

  /* 񱳴 빮ڷ ȯ ӽ ڿ ϸ,    */
  while((findp = strstr(str, tok)) != NULL) {
    for(i = 0; str != findp; str++, orgstr++, i++) buf[i] = *orgstr;
    buf[i] = '\0';
    qPuts(mode, buf);

    for(i = 0; i < strlen(orgtok); str++, orgstr++, i++) buf[i] = *orgstr;
    buf[i] = '\0';
    printf("<font color=red>%s</font>", buf);
  }
  qPuts(mode, orgstr);

  free(buf);
  free(freestr), free(freetok);
}

/********** Common Functions **************/

void print_feed(Q_Entry *sysp, SysConfig *sys, cDB *db, DBConfig *conf, int tArticle);
/*********** Main ***************************/
void cwb_feed(Q_Entry *sysp, SysConfig *sys, cDB *db, DBConfig *conf){
  char buf[0xFF];
  int  tArticle;

  /* DB Open */
  sprintf(buf, "%s/%s.%s", sys->DataDirectory, db->db, sys->DataExtension);
  DBOpen(buf, CWDB_READ);

  tArticle = main_index(db, conf);

  print_feed(sysp, sys, db, conf, tArticle);

  /* DB Close */
  DBClose();
}

void print_feed(Q_Entry *sysp, SysConfig *sys, cDB *db, DBConfig *conf, int tArticle){
  DBList head;

  int startNum,  endNum, pages;
  int startPage, endPage;

  int  i, status, virtualnum, filter;


  char *charset, *title;

  /* Ͱ  */
  if(!strcmp(conf->HtmlAccept, "true")) filter = 10;
  else if(!strcmp(conf->LinkTarget, "")) filter = 12;
  else filter = 13;

  /* Page   */
  if(tArticle == 0) pages = 0;
  else {
    int pof;

    if((tArticle % conf->ArticlesPerPage) == 0) pof = 0;
    else pof = 1;

    pages = (tArticle / conf->ArticlesPerPage) + pof;
  }

  /* ũ Page   */
  if(pages == 0) startPage = endPage = 0;
  else {
    int pof;

    if((db->page % atoi(sys->PageLinkNum)) == 0) pof = 0;
    else pof = 1;

    endPage   = ((db->page / atoi(sys->PageLinkNum)) + pof) * atoi(sys->PageLinkNum);
    startPage = endPage - (atoi(sys->PageLinkNum) - 1);
  }

  /*    */
  if(db->page > pages) db->page = pages;

  /*  Index ȣ  */
  if(db->page == 0) startNum = endNum = 0;
  else {
    startNum = (db->page - 1) * conf->ArticlesPerPage;
    endNum   = startNum + (conf->ArticlesPerPage - 1 );
  }

  if(qfValue(sysp, "Charset") != NULL) charset = qfValue(sysp, "Charset");
  else charset = "euc-kr";
  if(qfValue(sysp, "Title") != NULL) title = qfValue(sysp, "Title");
  else title = "CWB";

  printf("<?xml version='1.0' encoding='%s' ?>\n", charset);
  printf("<feed xmlns=\"http://www.w3.org/2005/Atom\">\n");
  printf("<title>%s</title>\n",title);
  printf("<link href=\"%s?db=%s\"></link>\n", sys->ProgramName, db->db);

  /* Ϻ  thread϶ ñ۵鶧  Խù ȣ
     ִ.  ̶ üù  ȣ Ͽ
     Ϸ ټֵ Virtual Num ɼ ־.
         Խù ǰų,  ñ ۼǸ
     Խù ȣ  Ѵٴ ̴.
     ׷, ñ϶ Խùȣ ϴ Ķ ̻ Ͽ
     ܰ  .
     忡 ڰ ̸ Ͽ,  ȣ ǥ ȣ
     ǥ Ѵ. */

  virtualnum = tArticle - ((db->page - 1) * conf->ArticlesPerPage);

  /* tIndex   0 ʵ尡 ԵǾ ִ */
  for(i = startNum; (i <= endNum) && (tIndex[i].num != 0); i++, virtualnum--){
    int h, m, s;
    char *text;
    status = getArticleInfo(tIndex[i].num, &head);

    printf("<entry>\n");

    /* Article */
    if(status != FALSE) {
      printf("<title>%s</title>\n", head.Subject);
      printf("<link>%s?db=%s&amp;mode=read&amp;num=%d&amp;page=%d&amp;ftype=%d&amp;fval=%s&amp;backdepth=1</link>\n",
        sys->ProgramName, db->db, tIndex[i].num, db->page, db->ftype, db->fencval);
      printf("<author><name>%s</name></author>\n", head.Name);
      sscanf(head.Date, "%4d%2d%2d%2d%2d%2d", &head.Year, &head.Mon, &head.Day,&h,&m,&s);
      printf("<updated>%02d-%02d-%02dT%02d:%02d:%02d+00:00</updated>\n", head.Year, head.Mon, head.Day,h,m,s);
      if((text = DBFetch("%d.Text", tIndex[i].num)) != NULL) {
        char *rtext;

        rtext = qStrReplace("sn", text, "]]", "&#93;&#93;");
        printf("<content><![CDATA[");
        qPuts(filter, rtext);
        printf("]]></content>\n");
      }
    }
    printf("</entry>\n");
    freeArticleInfo(&head);
  }
  printf("</feed>\n");
}
